var http = require('http');
var urlparse = require('url');
var mimeparse = require('mimeparse');

var respondWithStatus = function(res, status) {
  res.writeHead(status, {'Content-Type': 'text/plain'});
  res.end('\n');
};

exports.config = {
  match: function(path, mask) {
    return path.match(mask);
  },
  findMatch: function(path, urls) {
    for (var i in urls) {
      var url = urls[i];
      var match = exports.config.match(path, url.mask);
      if (match) {
        return [url, match];
      }
    }
    return false;
  }
};

exports.contentType = function(request) {
  return request.headers['content-type'];
};

function respondAfterAuth(request, response, method, match, auth, body) {
  if (method.understands) {
    var contentType = exports.contentType(request);
    for (var i in method.understands) {
      var mime = method.understands[i];
      if (mime == contentType) {
        respondAfterContentType(request, response, method, match, auth, body);
        break;
      }
      respondWithStatus(response, 415);
    }
  } else {
    respondAfterContentType(request, response, method, match, auth, body);
  }
}

function respondAfterContentType(request, response, method, match, auth, body) {
  var mime = null;
  var accept = request.headers['accept'];
  if (accept) {
    mime = mimeparse.bestMatch(method.provides, accept);
  }
  if (mime) {
    var args = {
      mime: mime,
      match: match[1].splice(1)
    };
    if (auth) {
      args.user = auth;
    }
    if (body) {
      if (request.headers["content-type"] === "application/json") {
        args.body = JSON.parse(body);
      } else {
        args.body = body;
      }
    }
    method.responder(request, response, args);
  } else {
    respondWithStatus(response, 406);
  }  
}

function onBody(reqOrRes, callback) {
  var body = "";
  reqOrRes.on('data', function(chunk) {
    body += chunk.toString('utf8');
  });
  reqOrRes.on('end', function() {
    callback(body);
  });
}

exports.createServer = function(urls) {
  var server = http.createServer(function (req, res) {
    var url = urlparse.parse(req.url);
    var match = exports.config.findMatch(url.pathname, urls);
    if (match === false) {
      respondWithStatus(res, 404);
    } else {
      var url = match[0];
      var method = url.methods[req.method];
      if (method) {
        
        onBody(req, function(body) {
          if (method.authorization) {
            if (req.headers.authorization) {
              var base64 = req.headers.authorization.match(/^Basic (.+)$/)[1];
              var userAndPassword = (new Buffer(base64, "base64")).toString().split(":", 2);
              method.authorization.authorizer(userAndPassword[0], userAndPassword[1], function(auth) {
                if (auth) {
                  respondAfterAuth(req, res, method, match, auth, body);
                } else {
                  respondWithStatus(res, 401);
                }
              });
            } else {
              if (method.authorization.optional) {
                respondAfterAuth(req, res, method, match, null, body);
              } else {
                respondWithStatus(res, 401);
              }
            }
          } else {
            respondAfterAuth(req, res, method, match, null, body);
          }
        });
        
      } else {
        respondWithStatus(res, 405);
      }
    }
  });
  return server;
};
